# -*- coding: utf-8 -*-
# pylint: disable=invalid-name

from itertools import chain
import types

from django.db import DatabaseError, transaction
from django import forms
from django.utils.functional import cached_property
from django.utils.module_loading import import_string


def yhdista_lomakkeet(
  LomakeA, LomakeB, *,
  avain_a=None,
  avain_b=None,
  tunnus=None,
  pakollinen_b=None,
  valita_parametrit=None,
):
  '''
  Yhdistää kaksi ModelForm-luokkaa silloin,
  kun A-luokasta on suora viittaus B-luokkaan

  Joko `avain_a` tai `avain_b` on pakollinen.

  Args:
    LomakeA, LomakeB: mallikohtaiset lomakeluokat
    avain_a: viittausavain LomakeA.Meta.model-luokan kohteelta
      LomakeB.Meta.model-luokan kohteelle
    avain_b: viittausavain LomakeB.Meta.model-luokan kohteelta
      LomakeA.Meta.model-luokan kohteelle
      (oletuksena haetaan viittauksen vastaluokan tiedoista)
    tunnus: yksikäsitteinen tunnus LomakeB-lomakkeelle
      (oletuksena ``avain_a`` tai ``avain_b``:n vastaviittaus)
    pakollinen_b: onko lomake B pakollinen, tarkistettava tieto?
      (oletuksena sama kuin A--B-viittauksen pakollisuus, jos annettu)
  '''

  if isinstance(LomakeA, str):
    LomakeA = import_string(LomakeA)
  if isinstance(LomakeB, str):
    LomakeB = import_string(LomakeB)

  # Muodosta avain_a, avain_b, kentta_a ja kentta_b.
  assert avain_a or avain_b
  kentta_a = LomakeA.Meta.model._meta.get_field(avain_a) if avain_a else None

  avain_b = avain_b or kentta_a.remote_field.name
  kentta_b = LomakeB.Meta.model._meta.get_field(avain_b)
  tunnus = tunnus or avain_a or kentta_b.remote_field.name

  # Mikäli A-avainta ei ole annettu,
  # B-lomake on oletuksena valinnainen.
  # Muutoin haetaan oletus vastaavan kentän tiedoista.
  if pakollinen_b is None:
    pakollinen_b = avain_a and (
      not LomakeA.Meta.model._meta.get_field(avain_a).null
    )

  class YhdistettyLomake(LomakeA):
    class Meta(LomakeA.Meta):
      pass

    def __init__(self, *args, prefix=None, **kwargs):
      lomake_kwargs = kwargs.pop(f'{tunnus}_kwargs', {})
      for param in (valita_parametrit or ()):
        try:
          lomake_kwargs[param] = kwargs[param]
        except KeyError:
          pass
      super().__init__(*args, prefix=prefix, **kwargs)
      # Poimitaan B-kohde joko `initial`-sanakirjasta tai A-kohteen tiedoista.
      kohde_b = self.initial.get(tunnus, None)
      if isinstance(kohde_b, LomakeB.Meta.model):
        pass
      elif kohde_b is not None:
        kohde_b = LomakeB.Meta.model.objects.get(pk=kohde_b)
      elif self.instance and avain_a:
        kohde_b = getattr(self.instance, avain_a, None)

      # Mikäli olemassaolevaa B-kohdetta ei löytynyt, luodaan uusi.
      if kohde_b is None:
        kohde_b = LomakeB.Meta.model()
        # Asetetaan linkki B-->A, jos mahdollista
        try:
          setattr(kohde_b, avain_b, self.instance)
        except (AttributeError, TypeError):
          pass
      assert isinstance(kohde_b, LomakeB.Meta.model), (
        f'Kohde B ei voi olla tyyppiä {type(kohde_b)} != {LomakeB.Meta.model}!'
      )

      lomake_b = LomakeB(
        instance=kohde_b,
        data=kwargs.get('data'),
        files=kwargs.get('files'),
        initial={
          avain.replace(tunnus + '-', '', 1): arvo
          for avain, arvo in self.initial.items()
          if avain.startswith(tunnus + '-') and avain != tunnus + '-'
        },
        prefix=f'{self.prefix}-{tunnus}' if self.prefix else tunnus,
        **lomake_kwargs
      )
      if avain_b in lomake_b.fields:
        lomake_b.fields[avain_b].disabled = True
        lomake_b.fields[avain_b].required = False
        lomake_b.fields[avain_b].widget = forms.HiddenInput()
      setattr(self, tunnus, lomake_b)
      # Jos B-viittaus saa olla tyhjä, asetetaan kaikki B-lomakkeen
      # kentät valinnaisiksi GET-pyynnöllä.
      # – Huomaa, että tämä koskee myös mahdollisten sisäkkäisten
      # lomakkeiden (C) kenttiä.
      # Lisäksi ohitetaan vimpainten `required`-määreen tulostus.
      if not pakollinen_b:
        for kentta in lomake_b:
          if not self.data:
            kentta.field.required = False
          kentta.field.widget.use_required_attribute = lambda initial: False
      # def __init__

    # def order_fields(self, field_order)
    # def __str__(self)
    # def __repr__(self)

    def __iter__(self):
      return chain(
        super().__iter__(),
        getattr(self, tunnus).__iter__(),
      )
      # def __iter__

    def __getitem__(self, item):
      if item.startswith(f'{tunnus}-'):
        return getattr(self, tunnus).__getitem__(
          item.partition(f'{tunnus}-')[2]
        )
      else:
        return super().__getitem__(item)
      # def __getitem__

    @property
    def errors(self):
      '''
      Lisää B-lomakkeen mahdolliset virheet silloin, kun
      B-viittaus ei saa olla tyhjä, tai B-lomaketta on muokattu.
      '''
      virheet = list(super().errors.items())
      if pakollinen_b or getattr(self, tunnus).has_changed():
        lomake_b = getattr(self, tunnus)
        for avain, arvo in list(lomake_b.errors.items()):
          virheet.append([
            '%s-%s' % (tunnus, avain), arvo
          ])
      return forms.utils.ErrorDict(virheet)
      # def errors

    def is_valid(self):
      '''
      Jos B-viittaus saa olla tyhjä eikä sitä ole muokattu,
      ei välitetä B-lomakkeen mahdollisesta epäkelpoisuudesta.
      '''
      return super().is_valid() \
      and (getattr(self, tunnus).is_valid() or (
        not pakollinen_b and not getattr(self, tunnus).has_changed()
      ))
      # def is_valid

    # def add_prefix(self, field_name)
    # def add_initial_prefix(self, field_name)

    def _html_output(self, *args, **kwargs):
      # pylint: disable=protected-access
      return super()._html_output(*args, **kwargs) \
      + getattr(self, tunnus)._html_output(*args, **kwargs)
      # def _html_output

    # def as_table(self)
    # def as_ul(self)
    # def as_p(self)
    # def non_field_errors(self)
    # def add_error(self, field, error)
    # def has_error(self, field, code=None)
    # def full_clean(self)
    # def _clean_fields(self)
    # def _clean_form(self)
    # def _post_clean(self)
    # def clean(self)

    def has_changed(self):
      return super().has_changed() \
      or getattr(self, tunnus).has_changed()
      # def has_changed

    @cached_property
    def changed_data(self):
      '''
      Palauta ylälomakkeen omien muutosten lisäksi
      liitoslomakkeen mahdolliset muutokset
      `tunnus`-etuliitteellä varustettuina.
      '''
      lomake = getattr(self, tunnus)
      return super().changed_data + [
        f'{tunnus}-{kentta}'
        for kentta in lomake.changed_data
      ]
      # def changed_data

    #@property
    #def media(self)

    #def is_multipart(self)
    #def hidden_fields(self)
    #def visible_fields(self)
    #def get_initial_for_field(self, field, field_name)


    # `in`

    def __contains__(self, key):
      if key.startswith(f'{tunnus}-'):
        key = key.partition(f'{tunnus}-')[2]
        if hasattr(getattr(self, tunnus), '__contains__') \
        and getattr(self, tunnus).__contains__(key):
          return True
        return key in getattr(self, tunnus).Meta.fields
        # if key.startswith
      if hasattr(super(), '__contains__') \
      and super().__contains__(key):
        return True
      else:
        return key in self.Meta.fields
      # def __contains__


    # ModelForm

    @transaction.atomic
    def _save_m2m(self):
      super()._save_m2m()
      # Jos viittaus A-->B voi olla tyhjä, tallennetaan B (vasta) nyt
      if not avain_a or self.Meta.model._meta.get_field(avain_a).null:

        # Haetaan A-kohde ja B-lomake
        kohde_a = self.instance
        lomake_b = getattr(self, tunnus)

        # Otetaan vanha B-kohde talteen
        vanha_kohde_b = getattr(kohde_a, avain_a) \
        if avain_a and hasattr(kohde_a, avain_a) else None

        if vanha_kohde_b == lomake_b.instance and vanha_kohde_b.pk:
          # Päivitetään olemassaoleva B
          try:
            lomake_b.save(commit=True)
          except (ValueError, DatabaseError):
            if pakollinen_b:
              raise
            # Säilytetään olemassaoleva tietue.
            vanha_kohde_b.refresh_from_db()
          # if vanha_kohde_b == lomake_b.instance and vanha_kohde_b.pk
        else:
          # Asetetaan linkki B-->A
          setattr(lomake_b.instance, avain_b, kohde_a)

          # Yritetään tallentaa B ja otetaan virhe kiinni
          try:
            kohde_b = lomake_b.save(commit=True)
          except (ValueError, DatabaseError):
            if pakollinen_b:
              raise
            kohde_b = vanha_kohde_b

          # Asetetaan tarvittaessa linkki A-->B ja tallennetaan A
          if avain_a:
            setattr(kohde_a, avain_a, kohde_b)
            kohde_a.save()

          # Poistetaan vanha kohde, jos viittaus katkesi
          if vanha_kohde_b and vanha_kohde_b.pk and vanha_kohde_b != kohde_b:
            vanha_kohde_b.delete()
          # if vanha_kohde_b != lomake_b.instance:
        # if ...null

      # def _save_m2m

    @transaction.atomic
    def save(self, commit=True):
      # Jos viittaus A-->B ei voi olla tyhjä, tallennetaan B ensin
      # (muussa tapauksessa B tallennetaan lopuksi `_save_m2m`-metodissa)
      if avain_a and not self.Meta.model._meta.get_field(avain_a).null:
        lomake_b = getattr(self, tunnus)
        kohde_b = lomake_b.save(commit=commit)
        setattr(self.instance, avain_a, kohde_b)

        # Kun `commit=False`, ja B on uusi, tallentamaton tietokantarivi:
        if not commit and kohde_b and not kohde_b.pk:
          # Vaihdetaan A:n tallennusmetodi kertaluontoisesti.
          vanha_save = self.instance.save
          def save(instance):
            # Hae tämänhetkinen B, tallenna se ja aseta uudelleen.
            kohde_b = getattr(instance, avain_a)
            kohde_b.save()
            # Päivitä B-lomakkeen rivi ja kutsu sen `save_m2m`-metodia.
            lomake_b.instance = kohde_b
            lomake_b.save_m2m()
            # Aseta B uudelleen A-kohteelle.
            setattr(instance, avain_a, kohde_b)
            # Palauta oletus-`save`-toteutus paikalleen ja kutsu sitä.
            instance.save = vanha_save
            instance.save()
            # def save
          self.instance.save = types.MethodType(save, self.instance)
          # if not commit and kohde_b and not kohde_b.pk

      return super().save(commit=commit)
      # def save

    # class YhdistettyLomake

  return YhdistettyLomake
  # def yhdista_lomakkeet



# -*- coding: utf-8 -*-
# pylint: disable=invalid-name

from itertools import chain

from django.db import transaction
from django.db.models import ProtectedError
from django import forms
from django.template import loader
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _


def lisaa_lomakesarja(
  LomakeA, LomakeB, *,
  avain_a=None,
  avain_b=None,
  tunnus=None,
  epasuora=False,
  lomakesarja_parametrit=None,
  valita_parametrit=None,
  **kwargs
):
  '''
  Yhdistää ModelForm-luokan ja toisesta ModelForm-luokasta
  muodostetun InlineFormSet-luokan silloin,
  kun B-luokasta on suora (ForeignKey) tai epäsuora (GenericForeignKey)
  viittaus A-luokkaan.

  Args:
    LomakeA, LomakeB: mallikohtaiset lomakeluokat
    avain_a, avain_b: viittaukset mallien välillä
    tunnus: yksikäsitteinen tunnus lomakesarjalle (oletus 'lomakesarja')
    epasuora: käytetään `generic_inlineformset_factory`-funktiota?
    lomakesarja_parametrit: parametrit, jotka asetetaan lomakesarjan määreiksi
    valita_parametrit: ulomman lomakkeen parametrit, jotka välitetään
      sellaisenaan lomakesarjalle
    *args, **kwargs: lisäparametrit ``inlineformset_factory``-funktiolle

  Returns:
    lomakeluokka
  '''
  # Pakotetaan ylimääräisten lomakkeiden määräksi nolla.
  kwargs['extra'] = 0

  assert tunnus or avain_a or avain_b, (
    'Joko `tunnus`, `avain_a` tai `avain_b` on määriteltävä.'
  )

  if epasuora:
    from django.contrib.contenttypes.forms import generic_inlineformset_factory
    lomakesarja = generic_inlineformset_factory(
      LomakeB.Meta.model,
      form=LomakeB,
      **kwargs
    )
  else:
    lomakesarja = forms.models.inlineformset_factory(
      LomakeA.Meta.model,
      LomakeB.Meta.model,
      form=LomakeB,
      **kwargs
    )

  # Käytetään tyhjän lomakkeen oletusarvoina
  # `initial`-datan ensimmäistä alkiota.
  class lomakesarja(lomakesarja):
    # pylint: disable=function-redefined
    def get_form_kwargs(self, index):
      if index is None and self.initial_extra:
        return {
          **super().get_form_kwargs(index), 'initial': self.initial_extra[0]
        }
      else:
        return super().get_form_kwargs(index)
      # def get_form_kwargs
    # class lomakesarja

  # Aseta mahdolliset lomakesarjan parametrit.
  for avain, arvo in (lomakesarja_parametrit or {}).items():
    setattr(lomakesarja, avain, arvo)

  # Lisää tarvittaessa oletusarvot HTML-piirtoa varten.
  if not hasattr(lomakesarja, 'label'):
    lomakesarja.label = (
      LomakeB.Meta.model._meta.verbose_name_plural
    ).capitalize()
  if not hasattr(lomakesarja, 'palikka'):
    lomakesarja.palikka = 'pumaska/lomakesarja_lomakekenttana.html'
  if not hasattr(lomakesarja, 'riviluokka'):
    lomakesarja.riviluokka = 'panel panel-default clearfix'
  if not hasattr(lomakesarja, 'lisaa_painike'):
    lomakesarja.lisaa_painike = _('Lisää %(malli)s') % {
      'malli': LomakeB.Meta.model._meta.verbose_name
    }
  if not hasattr(lomakesarja, 'poista_painike'):
    lomakesarja.poista_painike = _('Poista %(malli)s') % {
      'malli': LomakeB.Meta.model._meta.verbose_name
    }

  tunnus = tunnus or avain_a or (
    LomakeB.Meta.model._meta.get_field(avain_b).remote_field.name
  )

  class YhdistettyLomake(LomakeA):
    class Meta(LomakeA.Meta):
      pass

    def __init__(self, *args, prefix=None, **kwargs):
      lomakesarja_kwargs = kwargs.pop(f'{tunnus}_kwargs', {})
      for param in (valita_parametrit or ()):
        try:
          lomakesarja_kwargs[param] = kwargs[param]
        except KeyError:
          pass
      super().__init__(*args, prefix=prefix, **kwargs)
      initial = {
        avain.replace(tunnus + '-', '', 1): arvo
        for avain, arvo in self.initial.items()
        if avain.startswith(tunnus + '-') and avain != tunnus + '-'
      }
      setattr(self, tunnus, lomakesarja(
        data=kwargs.get('data'),
        files=kwargs.get('files'),
        instance=self.instance,
        initial=[initial] if initial else [],
        prefix=f'{self.prefix}-{tunnus}' if self.prefix else tunnus,
        **lomakesarja_kwargs,
      ))
      # def __init__

    # def order_fields(self, field_order)
    # def __str__(self)
    # def __repr__(self)

    def __iter__(self):
      return chain(
        super().__iter__(),
        *(lomake.__iter__() for lomake in getattr(self, tunnus).__iter__()),
      )
      # def __iter__

    def __getitem__(self, item):
      if item.startswith(f'{tunnus}-'):
        indeksi, _, item = item.partition(f'{tunnus}-')[2].partition('-')
        return getattr(self, tunnus).__getitem__(int(indeksi)).__getitem__(item)
      else:
        return super().__getitem__(item)
      # def __getitem__

    @property
    def errors(self):
      virheet = list(super().errors.items())
      lomakesarja = getattr(self, tunnus)
      for indeksi, lomake in enumerate(lomakesarja.forms):
        if lomake not in lomakesarja.deleted_forms:
          for avain, arvo in list(lomake.errors.items()):
            virheet.append([
              '%s-%d-%s' % (tunnus, indeksi, avain), arvo
            ])
      if any(lomakesarja.non_form_errors()):
        virheet.append([
          # Lisää lomakeriippumattomat virheet hallintolomakkeen kohdalle.
          tunnus + '-TOTAL_FORMS',
          lomakesarja.non_form_errors()
        ])
      return forms.utils.ErrorDict(virheet)
      # def errors

    def is_valid(self):
      return super().is_valid() \
      and getattr(self, tunnus).is_valid()
      # def is_valid

    # def add_prefix(self, field_name)
    # def add_initial_prefix(self, field_name)

    def _html_output(self, *args, **kwargs):
      # pylint: disable=protected-access
      return super()._html_output(*args, **kwargs) \
      + loader.get_template('pumaska/lomakesarja.html').render({
        'tunnus': tunnus,
        'lomakesarja': getattr(self, tunnus),
      })
      # def _html_output

    # def as_table(self)
    # def as_ul(self)
    # def as_p(self)
    # def non_field_errors(self)
    # def add_error(self, field, error)
    # def has_error(self, field, code=None)
    # def full_clean(self)
    # def _clean_fields(self)
    # def _clean_form(self)
    # def _post_clean(self)
    # def clean(self)

    def has_changed(self):
      return super().has_changed() \
      or getattr(self, tunnus).has_changed()
      # def has_changed

    @cached_property
    def changed_data(self):
      '''
      Palauta ylälomakkeen omien muutosten lisäksi
      liitoslomakkeiden mahdolliset muutokset
      lomakesarjan määrittämillä, lomakekohtaisella
      etuliitteillä varustettuina.
      '''
      lomakesarja = getattr(self, tunnus)
      # Muodosta lomakekohtainen kentän etuliite poistamalla
      # liitetyn lomakkeen `prefixin` alusta
      # käsillä olevan (ylä-) lomakkeen oma `prefix` ja välimerkki -.
      lomakekohtainen_tunnus = (
        lambda lomake: lomake.prefix.replace(self.prefix + "-", "", 1)
      ) if self.prefix else lambda lomake: lomake.prefix
      return super().changed_data + sum([[
        f'{lomakekohtainen_tunnus(lomake)}-{kentta}'
        for kentta in lomake.changed_data
      ] for lomake in lomakesarja], [])
      # def changed_data

    #@property
    #def media(self)

    #def is_multipart(self)
    #def hidden_fields(self)
    #def visible_fields(self)
    #def get_initial_for_field(self, field, field_name)


    # `in`

    def __contains__(self, key):
      if key.startswith(f'{tunnus}-'):
        indeksi, _, key = key.partition(f'{tunnus}-')[2].partition('-')
        if hasattr(
          getattr(self, tunnus).__getitem__(int(indeksi)),
          '__contains__'
        ) and getattr(self, tunnus).__getitem__(int(indeksi)).__contains__(key):
          return True
        return key in getattr(self, tunnus).__getitem__(
          int(indeksi)
        ).Meta.fields
        # if key.startswith
      if hasattr(super(), '__contains__') \
      and super().__contains__(key):
        return True
      else:
        return key in self.fields
      # def __contains__


    # ModelForm

    @transaction.atomic
    def save(self, commit=True):
      '''
      Tallennetaan atomaarisena tietokantaoperaationa.
      '''
      return super().save(commit=commit)
      # def save

    def _save_m2m(self):
      '''
      Tallennetaan M2M-kohteet `super`-toteutuksen mukaisesti.
      Tämän jälkeen tallennetaan lomakesarja.
      '''
      super()._save_m2m()
      lomakesarja = getattr(self, tunnus)
      lomakesarja.instance = self.instance
      try:
        lomakesarja.save(commit=True)
      except ProtectedError as exc:
        virheteksti = _(
          'Rivin poisto epäonnistui:'
          ' suojattuja, riippuvia %(malli)s-kohteita.'
        ) % {'malli': exc.protected_objects.model._meta.verbose_name}
        # pylint: disable=protected-access
        lomakesarja._non_form_errors.append(forms.ValidationError(
          virheteksti, code='protect'
        ))
        raise forms.ValidationError(exc)
      # def _save_m2m

    # class YhdistettyLomake

  return YhdistettyLomake
  # def lisaa_lomakesarja



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

setup(
  setup_requires='git-versiointi',
  name='django-pumaska',
  description='Sisäkkäisten lomakkeiden ja -sarjojen käsittely',
  url='https://github.com/an7oine/django-pumaska',
  author='Antti Hautaniemi',
  author_email='antti.hautaniemi@pispalanit.fi',
  packages=find_packages(),
  include_package_data=True,
  entry_points={
    'django.sovellus': [
      'pumaska = pumaska',
    ],
  },
  zip_safe=False,
)



# -*- coding: utf-8 -*-
# pylint: disable=invalid-name

import contextlib
from decimal import Decimal
import itertools
from unittest.mock import patch

from django.db import models
from django import forms
from django.test import SimpleTestCase

from pumaska import nido


class Asiakas(models.Model):
  nimi = models.CharField(max_length=255)

class Asiakasosoite(models.Model):
  asiakas = models.ForeignKey(Asiakas, on_delete=models.CASCADE)
  osoite = models.CharField(max_length=255)

class Paamies(models.Model):
  nimi = models.CharField(max_length=255)

class Lasku(models.Model):
  asiakas = models.ForeignKey(Asiakas, on_delete=models.CASCADE)
  paamies = models.ForeignKey(Paamies, on_delete=models.CASCADE)
  numero = models.IntegerField()

class Rivi(models.Model):
  lasku = models.ForeignKey(Lasku, on_delete=models.CASCADE)
  summa = models.DecimalField(max_digits=11, decimal_places=2)


class Testi(SimpleTestCase):

  @contextlib.contextmanager
  def varmista_tietueiden_tallennus(self, tietueet):
    '''
    Varmista, että suoritettava koodi tallentaa tietokantaan
    täsmälleen annettuja sanakirjoja vastaavat tietueet
    annetussa järjestyksessä.
    '''
    tietueet = iter(tietueet)
    def tallenna(self2):
      tietue = {'__class__': type(self2).__name__, **{
        k: v for k, v in self2.__dict__.items()
        if k not in ('id', '_state')
        and not k.endswith('_id')
        and not isinstance(v, models.Model)
      }}
      self.assertEqual(tietue, next(tietueet))
      # def tallenna
    with patch('django.db.transaction.Atomic.__enter__'):
      with patch('django.db.transaction.Atomic.__exit__'):
        with patch('django.db.models.base.Model.save', tallenna):
          yield
    self.assertRaises(StopIteration, next, tietueet)
    # def varmista_tietueiden_tallennus

  def testaa_html(self):
    'Toimiiko HTML-sivun muodostus liitoslomakkeen mukaan?'
    # pylint: disable=line-too-long
    html = str(nido(
      forms.modelform_factory(Asiakas, fields=['nimi']),
      forms.modelform_factory(Asiakasosoite, fields=['osoite']),
      nido(
        nido(
          forms.modelform_factory(Lasku, fields=['numero']),
          forms.modelform_factory(Paamies, fields=['nimi']),
          avain_a='paamies'
        ),
        forms.modelform_factory(Rivi, fields=['summa']),
        avain_b='lasku'
      ),
    )())
    for name in map('-'.join, itertools.chain.from_iterable(itertools.starmap(itertools.product, (
      (('asiakasosoite', 'lasku', ), ('TOTAL_FORMS', 'INITIAL_FORMS', 'MIN_NUM_FORMS', 'MAX_NUM_FORMS', )),
      (('asiakasosoite', ), ('__prefix__', ), ('id', 'asiakas', 'osoite', 'DELETE', ), ),
      (('lasku', ), ('__prefix__', ), ('id', 'asiakas', 'numero', 'DELETE', )),
      (('lasku', ), ('__prefix__', ), ('rivi', ), ('TOTAL_FORMS', 'INITIAL_FORMS', 'MIN_NUM_FORMS', 'MAX_NUM_FORMS', )),
      (('lasku', ), ('__prefix__', ), ('rivi', ), ('__prefix__', ), ('id', 'lasku', 'summa', 'DELETE', )),
    )))):
      self.assertIn(f' name="{name}" ', html)
    # def testaa_html

  def testaa_liitospuu(self):
    'Toimiiko monihaaraisen lomakepuun tallennus?'

    lomake = nido(
      forms.modelform_factory(Asiakas, fields=['nimi']),
      forms.modelform_factory(Asiakasosoite, fields=['osoite']),
      nido(
        nido(
          forms.modelform_factory(Lasku, fields=['numero']),
          forms.modelform_factory(Paamies, fields=['nimi']),
        ),
        forms.modelform_factory(Rivi, fields=['summa']),
      ),
    )(data={
      'nimi': 'Testimies Matti',
      'asiakasosoite-TOTAL_FORMS': '2',
      'asiakasosoite-INITIAL_FORMS': '0',
      'asiakasosoite-MIN_NUM_FORMS': '0',
      'asiakasosoite-MAX_NUM_FORMS': '1000',
      'asiakasosoite-0-osoite': 'Testitie 2',
      'asiakasosoite-1-osoite': 'Testitie 3',
      'lasku-TOTAL_FORMS': '1',
      'lasku-INITIAL_FORMS': '0',
      'lasku-MIN_NUM_FORMS': '0',
      'lasku-MAX_NUM_FORMS': '1000',
      'lasku-0-numero': '12345',
      'lasku-0-paamies-nimi': 'Palapallon Turrukkamatkat',
      'lasku-0-rivi-TOTAL_FORMS': '2',
      'lasku-0-rivi-INITIAL_FORMS': '0',
      'lasku-0-rivi-MIN_NUM_FORMS': '0',
      'lasku-0-rivi-MAX_NUM_FORMS': '1000',
      'lasku-0-rivi-0-summa': '123.45',
      'lasku-0-rivi-1-summa': '678.90',
    })
    self.assertTrue(lomake.is_valid())

    with self.varmista_tietueiden_tallennus((
      {'__class__': 'Asiakas', 'nimi': 'Testimies Matti'},
      {'__class__': 'Asiakasosoite', 'osoite': 'Testitie 2'},
      {'__class__': 'Asiakasosoite', 'osoite': 'Testitie 3'},
      {'__class__': 'Paamies', 'nimi': 'Palapallon Turrukkamatkat'},
      {'__class__': 'Lasku', 'numero': 12345},
      {'__class__': 'Rivi', 'summa': Decimal('123.45'),},
      {'__class__': 'Rivi', 'summa': Decimal('678.90'),},
    )):
      lomake.save()
    # def testaa_tallennus

  def testaa_koristeena(self):
    'Toimiiko @nido-koristeen kautta luotu lomake oikein?'

    class Asiakasosoitelomake(forms.ModelForm):
      class Meta:
        model = Asiakasosoite
        fields = ['osoite']
    class Laskulomake(forms.ModelForm):
      class Meta:
        model = Lasku
        fields = ['numero']

    @nido(Asiakasosoitelomake)
    @nido(Laskulomake)
    class Asiakaslomake(forms.ModelForm):
      class Meta:
        model = Asiakas
        fields = ['nimi']

    lomake = Asiakaslomake(data={
      'nimi': 'Testimies Matti',
      'asiakasosoite-TOTAL_FORMS': '1',
      'asiakasosoite-INITIAL_FORMS': '0',
      'asiakasosoite-MIN_NUM_FORMS': '0',
      'asiakasosoite-MAX_NUM_FORMS': '1000',
      'asiakasosoite-0-osoite': 'Testitie 2',
      'lasku-TOTAL_FORMS': '2',
      'lasku-INITIAL_FORMS': '0',
      'lasku-MIN_NUM_FORMS': '0',
      'lasku-MAX_NUM_FORMS': '1000',
      'lasku-0-numero': '12345',
      'lasku-1-numero': '15445',
    })
    self.assertTrue(lomake.is_valid())
    with self.varmista_tietueiden_tallennus((
      {'__class__': 'Asiakas', 'nimi': 'Testimies Matti'},
      {'__class__': 'Lasku', 'numero': 12345},
      {'__class__': 'Lasku', 'numero': 15445},
      {'__class__': 'Asiakasosoite', 'osoite': 'Testitie 2'},
    )):
      lomake.save()
    # def testaa_koristeena

  # class Testi
